from json_database import JsonDatabase
from mock_mycroft_backend.configuration import CONFIGURATION


class Metric:
    def __init__(self, metric_id, metric_type, json=""):
        self.metric_id = metric_id
        self.metric_type = metric_type
        self.json = json


class JsonMetricDatabase(JsonDatabase):
    def __init__(self, path=CONFIGURATION["metrics_db"]):
        super().__init__("metrics", path)

    def add_metric(self, metric_type=None, json="{}"):
        metric_id = self.total_metrics() + 1
        metric = Metric(metric_id, metric_type, json)
        self.add_item(metric)

    def total_metrics(self):
        return len(self)

    def __enter__(self):
        """ Context handler """
        return self

    def __exit__(self, _type, value, traceback):
        """ Commits changes and Closes the session """
        try:
            self.commit()
        except Exception as e:
            print(e)




# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from functools import wraps
from flask import make_response


def add_response_headers(headers=None):
    """This decorator adds the headers passed in to the response"""
    headers = headers or {}

    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            resp = make_response(f(*args, **kwargs))
            h = resp.headers
            for header, value in headers.items():
                h[header] = value
            return resp

        return decorated_function

    return decorator


def noindex(f):
    """This decorator passes X-Robots-Tag: noindex"""
    return add_response_headers({'X-Robots-Tag': 'noindex'})(f)




from mock_mycroft_backend.backend import start_backend
from mock_mycroft_backend.configuration import CONFIGURATION


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--flask-port", help="Mock backend port number",
                        default=CONFIGURATION["backend_port"])
    parser.add_argument("--flask-host", help="Mock backend host",
                        default="127.0.0.1")
    args = parser.parse_args()
    start_backend(args.flask_port, args.flask_host)


if __name__ == "__main__":
    main()



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
from requests import get
from mock_mycroft_backend.backend.decorators import noindex
from mock_mycroft_backend.utils import nice_json
from flask import send_file, request


def build_response(audio_file, visimes=None):
    if visimes is not None:
        with open(audio_file, "rb") as f:
            audio_data = f.read()
        encoded_audio = base64.b64encode(audio_data)
        res = {
            "audio_base64": encoded_audio.decode("utf-8"),
            "visimes": visimes
        }
        return nice_json(res)
    else:
        return send_file(
            audio_file,
            mimetype="audio/wav")


def get_tts_routes(app):
    @app.route("/synthesize/mimic2/<voice>/<lang>", methods=['GET'])
    @noindex
    def mimic2_proxy(voice, lang):
        # TODO cache results to save calls to mycroft.ai
        return get("https://mimic-api.mycroft.ai/synthesize", params=request.args).content

    return app



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask import Flask
from flask_mail import Mail
from flask_sslify import SSLify
from mock_mycroft_backend.configuration import CONFIGURATION


API_VERSION = CONFIGURATION["api_version"]


def create_app():
    app = Flask(__name__)

    mail = Mail(app)
    if CONFIGURATION["ssl"]:
        sslify = SSLify(app)

    from mock_mycroft_backend.utils import nice_json
    from mock_mycroft_backend.backend.decorators import noindex
    from mock_mycroft_backend.backend.auth import get_auth_routes
    from mock_mycroft_backend.backend.device import get_device_routes
    from mock_mycroft_backend.backend.stt import get_stt_routes
    from mock_mycroft_backend.backend.tts import get_tts_routes

    app = get_auth_routes(app)
    app = get_device_routes(app, mail)
    app = get_stt_routes(app)
    app = get_tts_routes(app)

    @app.route("/", methods=['GET'])
    @noindex
    def hello():
        return nice_json({
            "message": "Welcome to Mock Mycroft Backend",
            "author": "JarbasAI"
        })

    return app


def start_backend(port=CONFIGURATION["backend_port"], host="0.0.0.0"):
    if CONFIGURATION["ssl"]:
        import ssl
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        context.load_cert_chain(CONFIGURATION["ssl_cert"],
                                CONFIGURATION["ssl_key"])

        app = create_app()
        app.run(port=port, ssl_context=context,
                use_reloader=False, host=host)
    else:
        app = create_app()
        app.run(port=port, use_reloader=False, host=host)
    return app


if __name__ == "__main__":
    start_backend()



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import subprocess
import tarfile
from os.path import exists, dirname
from threading import Thread

_running_downloads = {}


def _get_download_tmp(dest):
    tmp_base = dest + '.part'
    if not exists(tmp_base):
        return tmp_base
    else:
        i = 1
        while (True):
            tmp = tmp_base + '.' + str(i)
            if not exists(tmp):
                return tmp
            else:
                i += 1


class Downloader(Thread):
    """
        Downloader is a thread based downloader instance when instanciated
        it will download the provided url to a file on disk.

        When the download is complete or failed the `.done` property will
        be set to true and the `.status` will indicate the status code.
        200 = Success.

        Args:
            url:            Url to download
            dest:           Path to save data to
            complet_action: Function to run when download is complete.
                            `func(dest)`
    """

    def __init__(self, url, dest, complete_action=None, header=None):
        super(Downloader, self).__init__()
        self.url = url
        self.dest = dest
        self.complete_action = complete_action
        self.status = None
        self.done = False
        self._abort = False
        self.header = header

        # Create directories as needed
        if not exists(dirname(dest)):
            os.makedirs(dirname(dest))

        #  Start thread
        self.daemon = True
        self.start()

    def perform_download(self, dest):

        cmd = ['wget', '-c', self.url, '-O', dest,
               '--tries=20', '--read-timeout=5']
        if self.header:
            cmd += ['--header={}'.format(self.header)]
        return subprocess.call(cmd)

    def run(self):
        """
            Does the actual download.
        """
        tmp = _get_download_tmp(self.dest)
        self.status = self.perform_download(tmp)

        if not self._abort and self.status == 0:
            self.finalize(tmp)
        else:
            self.cleanup(tmp)
        self.done = True
        arg_hash = hash(self.url + self.dest)

        #  Remove from list of currently running downloads
        if arg_hash in _running_downloads:
            _running_downloads.pop(arg_hash)

    def finalize(self, tmp):
        """
            Move the .part file to the final destination and perform any
            actions that should be performed at completion.
        """
        os.rename(tmp, self.dest)
        if self.complete_action:
            self.complete_action(self.dest)

    def cleanup(self, tmp):
        """
            Cleanup after download attempt
        """
        if exists(tmp):
            os.remove(self.dest + '.part')
        if self.status == 200:
            self.status = -1

    def abort(self):
        """
            Abort download process
        """
        self._abort = True


def download(url, dest, complete_action=None, header=None):
    global _running_downloads
    arg_hash = hash(url + dest)
    if arg_hash not in _running_downloads:
        _running_downloads[arg_hash] = Downloader(url, dest, complete_action,
                                                  header)
    return _running_downloads[arg_hash]


def untar(file_path, dest, remove=False):
    if not exists(file_path):
        raise AssertionError("file does not exist")
    if not file_path.endswith(".tar.gz") or not file_path.endswith(".tar.bz2"):
        raise AssertionError("invalid file format")
    with tarfile.open(file_path) as tar:
        tar.extractall(dest)
    if remove:
        os.remove(file_path)



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from flask import make_response
from subprocess import check_output
import random
import json
from os.path import exists, join, dirname
import requests


def root_dir():
    """ Returns root directory for this project """
    return dirname(dirname(__file__))


def generate_code():
    k = ""
    while len(k) < 6:
        k += random.choice(["A", "B", "C", "D", "E", "F", "G", "H", "I",
                            "J", "K", "L", "M", "N", "O", "P", "Q", "R",
                            "S", "T", "U", "Y", "V", "X", "W", "Z", "0",
                            "1", "2", "3", "4", "5", "6", "7", "8", "9"])
    return k.upper()


def nice_json(arg):
    response = make_response(json.dumps(arg, sort_keys=True, indent=4))
    response.headers['Content-type'] = "application/json"
    return response


def geo_locate(ip):
    if ip in ["0.0.0.0", "127.0.0.1"]:
        ip = str(check_output(['hostname', '--all-ip-addresses'])).split(" ")[0][2:]
    data = requests.get("http://ip-api.com/json/" + ip,
                        params={"fields": "status,country,countryCode,region,regionName,city,lat,lon,timezone,query"})\
        .json()
    return location_dict(
        city=data["city"],
        region_code=data["region"],
        country_code=data["countryCode"],
        country_name=data["country"],
        region=data["regionName"],
        longitude=data["lon"],
        latitude=data["lat"],
        timezone=data["timezone"]
    )


def location_dict(city="", region_code="", country_code="",
                  country_name="", region="", longitude=0, latitude=0,
                  timezone="", **kwargs):
    region_data = {"code": region_code, "name": region,
                   "country": {
                       "code": country_code,
                       "name": country_name}}
    city_data = {"code": city, "name": city,
                 "state": region_data,
                 "region": region_data}
    timezone_data = {"code": timezone, "name": timezone,
                     "dstOffset": 3600000,
                     "offset": -21600000}
    coordinate_data = {"latitude": float(latitude),
                       "longitude": float(longitude)}
    return {"city": city_data,
            "coordinate": coordinate_data,
            "timezone": timezone_data}



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mock_mycroft_backend.backend import API_VERSION
from mock_mycroft_backend.utils import nice_json
from mock_mycroft_backend.backend.decorators import noindex
from flask import request
import time


def get_auth_routes(app):
    @app.route("/" + API_VERSION + "/pair/<code>/<uuid>/<name>/<mail>",
               methods=['PUT'])
    @noindex
    def pair(code, uuid, name, mail):
        # auto - pair
        return nice_json({"paired": True})

    @app.route("/" + API_VERSION + "/auth/token", methods=['GET'])
    @noindex
    def token():
        token = request.headers.get('Authorization', '').replace("Bearer ", "")
        device = {"uuid": "AnonDevice",
                  "expires_at": time.time() + 999999999999999999,
                  "accessToken": token,
                  "refreshToken": token}
        return nice_json(device)

    return app



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from tempfile import NamedTemporaryFile
import json
from flask import request
from speech_recognition import Recognizer, AudioFile

from mock_mycroft_backend.backend import API_VERSION
from mock_mycroft_backend.configuration import CONFIGURATION
from mock_mycroft_backend.backend.decorators import noindex
from speech2text import STTFactory

recognizer = Recognizer()
engine = STTFactory.create(CONFIGURATION["stt"])


def get_stt_routes(app):
    @app.route("/" + API_VERSION + "/stt", methods=['POST'])
    @noindex
    def stt():
        flac_audio = request.data
        lang = str(request.args.get("lang", "en-us"))
        with NamedTemporaryFile() as fp:
            fp.write(flac_audio)
            with AudioFile(fp.name) as source:
                audio = recognizer.record(source)  # read the entire audio file

            utterance = engine.execute(audio, language=lang)
        return json.dumps([utterance])

    return app



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from os.path import join, expanduser, isdir, exists
from os import makedirs
from json_database import JsonStorage

# DATABASE
DATA_PATH = join(expanduser("~/.mycroft/mock_backend"))
if not isdir(DATA_PATH):
    makedirs(DATA_PATH)
METRICS_DB = join(DATA_PATH, "metrics.json")
CONFIG_PATH = join(DATA_PATH, "mock_backend.conf")

# SSL
SSL = False
SSL_CERT = join(DATA_PATH, "mock_backend.crt")
SSL_KEY = join(DATA_PATH, "mock_backend.key")

# MOCK BACKEND
API_VERSION = "v1"
BACKEND_PORT = 6712

# EMAIL

MAIL_SERVER = 'smtp.googlemail.com'
MAIL_PORT = 465
MAIL_USE_TLS = False
MAIL_USE_SSL = True
MAIL_USERNAME = "xxx@gmail.com"
MAIL_PASSWORD = "xxx"
MAIL_DEFAULT_SENDER = MAIL_USERNAME

RECEIVING_EMAIL = MAIL_USERNAME

STT_CONFIG = {"module": "google",
              "google": {}}

LANG = "en-us"

OVERRIDE_LOCATION = False
GEOLOCATE_IP = False
DEFAULT_LOCATION = {
    "city": {
        "code": "Lawrence",
        "name": "Lawrence",
        "state": {
            "code": "KS",
            "name": "Kansas",
            "country": {
                "code": "US",
                "name": "United States"
            }
        }
    },
    "coordinate": {
        "latitude": 38.971669,
        "longitude": -95.23525
    },
    "timezone": {
        "code": "America/Chicago",
        "name": "Central Standard Time",
        "dstOffset": 3600000,
        "offset": -21600000
    }
}


def default_conf():
    default = {
        "lang": LANG,
        "stt": STT_CONFIG,
        "backend_port": BACKEND_PORT,
        "ssl": SSL,
        "ssl_cert": SSL_CERT,
        "ssl_key": SSL_KEY,
        "mail_user": MAIL_USERNAME,
        "mail_password": MAIL_PASSWORD,
        "mail_server": MAIL_SERVER,
        "mail_port": MAIL_PORT,
        "default_location": DEFAULT_LOCATION,
        "geolocate": GEOLOCATE_IP,
        "override_location": OVERRIDE_LOCATION,
        "data_dir": DATA_PATH,
        "metrics_db": METRICS_DB,
        "api_version": API_VERSION,
        "email": RECEIVING_EMAIL,
        "data_path": DATA_PATH
    }
    config = JsonStorage(CONFIG_PATH)
    for k in default:
        config[k] = default[k]
    return config


if not exists(CONFIG_PATH):
    CONFIGURATION = default_conf()
    CONFIGURATION.store()
else:
    CONFIGURATION = JsonStorage(CONFIG_PATH)



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from flask_mail import Message
from flask import request
from mock_mycroft_backend.utils import geo_locate, generate_code, nice_json
from mock_mycroft_backend.configuration import CONFIGURATION
from mock_mycroft_backend.backend import API_VERSION
from mock_mycroft_backend.backend.decorators import noindex
from mock_mycroft_backend.database.metrics import JsonMetricDatabase
import time
import json


def get_device_routes(app, mail_sender):
    @app.route("/v1/device/<uuid>/settingsMeta", methods=['PUT'])
    def settingsmeta(uuid):
        return nice_json({"success": True, "uuid": uuid})

    @app.route("/v1/device/<uuid>/skill/settings", methods=['GET'])
    def skill_settings(uuid):
        return nice_json({"backend_disabled": True})

    @app.route("/v1/device/<uuid>/skillJson", methods=['PUT'])
    def skill_json(uuid):
        return nice_json({"backend_disabled": True})

    @app.route("/" + API_VERSION + "/device/<uuid>/location", methods=['GET'])
    @noindex
    def location(uuid):
        if not request.headers.getlist("X-Forwarded-For"):
            ip = request.remote_addr
        else:
            # TODO http://esd.io/blog/flask-apps-heroku-real-ip-spoofing.html
            ip = request.headers.getlist("X-Forwarded-For")[0]
        if CONFIGURATION["override_location"]:
            new_location = CONFIGURATION["default_location"]
        elif CONFIGURATION["geolocate"]:
            new_location = geo_locate(ip)
        else:
            new_location = {}
        return nice_json(new_location)

    @app.route("/" + API_VERSION + "/device/<uuid>/setting", methods=['GET'])
    @noindex
    def setting(uuid=""):
        result = {}
        return nice_json(result)

    @app.route("/" + API_VERSION + "/device/<uuid>", methods=['PATCH', 'GET'])
    @noindex
    def get_uuid(uuid):
        device = {"uuid": "AnonDevice",
                  "expires_at": time.time() + 72000,
                  "accessToken": uuid,
                  "refreshToken": uuid}
        return nice_json(device)

    @app.route("/" + API_VERSION + "/device/code", methods=['GET'])
    @noindex
    def code():
        uuid = request.args["state"]
        code = generate_code()
        result = {"code": code, "uuid": uuid}
        return nice_json(result)

    @app.route("/" + API_VERSION + "/device/", methods=['GET'])
    @noindex
    def device():
        token = request.headers.get('Authorization', '').replace("Bearer ", "")
        device = {"uuid": "AnonDevice",
                  "expires_at": time.time() + 72000,
                  "accessToken": token,
                  "refreshToken": token}
        return nice_json(device)

    @app.route("/" + API_VERSION + "/device/activate", methods=['POST'])
    @noindex
    def activate():
        token = request.json["state"]
        device = {"uuid": "AnonDevice",
                  "expires_at": time.time() + 72000,
                  "accessToken": token,
                  "refreshToken": token}
        return nice_json(device)

    @app.route("/" + API_VERSION + "/device/<uuid>/message", methods=['PUT'])
    @noindex
    def send_mail(uuid=""):
        data = request.json
        message = data["body"]
        subject = data["title"]
        msg = Message(recipients=[CONFIGURATION["email"]],
                      body=message,
                      subject=subject,
                      sender=data["sender"])
        mail_sender.send(msg)

    @app.route("/" + API_VERSION + "/device/<uuid>/metric/<name>",
               methods=['POST'])
    @noindex
    def metric(uuid="", name=""):
        data = request.json
        with JsonMetricDatabase() as db:
            db.add_metric(name, json.dumps(data))

    @app.route("/" + API_VERSION + "/device/<uuid>/subscription",
               methods=['GET'])
    @noindex
    def subscription_type(uuid=""):
        sub_type = "free"
        subscription = {"@type": sub_type}
        return nice_json(subscription)

    @app.route("/" + API_VERSION + "/device/<uuid>/voice", methods=['GET'])
    @noindex
    def get_subscriber_voice_url(uuid=""):
        arch = request.args["arch"]
        return nice_json({"link": "",
                          "arch": arch})

    return app



from setuptools import setup
import os


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


extra_files = package_files('mock_mycroft_backend')

setup(
    name='mock-mycroft-backend',
    version='0.2.1',
    packages=['mock_mycroft_backend',
              'mock_mycroft_backend.backend',
              'mock_mycroft_backend.database'],
    install_requires=['Flask>=0.12', 'requests>=2.2.1', 'Flask-Mail',
                      'speech2text', 'pyOpenSSL', "flask_sslify",
                      "json_database"],
    package_data={'': extra_files},
    include_package_data=True,
    url='https://github.com/OpenJarbas/mock-backend',
    license='Apache-2.0',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='mock mycroft backend',
    entry_points={
        'console_scripts': [
            'mock-mycroft-backend=mock_mycroft_backend.__main__:main'
        ]
    }
)
